using static System.Console;
using static System.Math;

using Cairo;
using Gdk;
using Gtk;
using Rectangle = Cairo.Rectangle;

class View : DrawingArea {
    const int SQUARE_SIZE = 100;
    const int MARGIN = SQUARE_SIZE / 2;
    
    Game game;
    Player ai;
    
    void init() {
        QueueDraw();     // draw the initial position
        game.changed += QueueDraw;
    }
    
    public View(Player ai) {
        game = new Game();
        this.ai = ai;
        init();
        AddEvents((int) EventMask.ButtonPressMask);    // ask to receive button press events
    }

    void drawLine(Context c, double x1, double y1, double x2, double y2) {
        c.MoveTo(x1, y1);
        c.LineTo(x2, y2);
        c.Stroke();
    }

    void drawX(Context c, Rectangle r) {
        drawLine(c, r.X, r.Y, r.X + r.Width, r.Y + r.Height);
        drawLine(c, r.X + r.Width, r.Y, r.X, r.Y + r.Height);
    }
    
    void drawO(Context c, Rectangle r) {
        c.Arc(r.X + r.Width / 2, r.Y + r.Height / 2, r.Width / 2, 0.0, 2 * PI);
        c.Stroke();
    }
    
    protected override bool OnButtonPressEvent(EventButton ev) {
        if (ev.Type == EventType.ButtonPress)
            if (game.winner() != null) {
                game = new Game();
                init();     // start a new game
            }
            else {
                int x = ((int) (ev.X - MARGIN)) / SQUARE_SIZE;
                int y = ((int) (ev.Y - MARGIN)) / SQUARE_SIZE;
                game.move(new Move(x, y));

                if (game.winner() == null) {
                    game.move(ai.move(game));
                }
            }

        return true;
    }

    protected override bool OnDrawn(Context c) {
        // Set a transformation matrix so that squares are 1 unit tall/wide, with a
        // MARGIN around the drawing area.
        c.Translate(MARGIN, MARGIN);
        c.Scale(SQUARE_SIZE, SQUARE_SIZE);
        
        // Adjust the line width to compensate for the scaling tranformation.
        c.LineWidth /= SQUARE_SIZE;
        
        if (game.won > 0) {
            c.SetSourceRGB(0.0, 1.0, 0.0);  // green
            for (int i = 0 ; i < 3 ; ++i) {
                c.Rectangle(game.win_x + i * game.win_dx,
                                        game.win_y + i * game.win_dy,
                                        1, 1);
                c.Fill();
            }
            c.SetSourceRGB(0.0, 0.0, 0.0);  // black
        }
        
        // Draw the tic tac toe grid.
        for (int i = 1 ; i <= 2 ; ++i) {
            drawLine(c, i, 0, i, 3);
            drawLine(c, 0, i, 3, i);
        }
        
        // Draw the X's and O's.
        for (int x = 0 ; x < 3 ; ++x)
            for (int y = 0 ; y < 3 ; ++y) {
                Rectangle rect = new Rectangle(x + 0.1, y + 0.1, 0.8, 0.8);
                
                switch (game[x, y]) {
                    case 1: drawX(c, rect); break;
                    case 2: drawO(c, rect); break;
                }
            }
        return true;
    }
}

class Frame : Gtk.Window {
    public Frame(Player ai) : base("tic tac toe") {
        Resize(400, 400);
        Add(new View(ai));
    }
    
    protected override bool OnDeleteEvent(Event e) {
        Application.Quit();
        return true;
    }
}

class MainClass {
    static Player player(string name) => name switch {
        "random" => new RandomPlayer(),
        "minimax" => new MinimaxPlayer(),
        _ => throw new Exception("unknown player")
    };

    static void tournament(int games, string name1, string name2) {
        string[] names = { "draw", name1, name2 };
        Player?[] players = { null, player(name1), player(name2) };
        int[] won = new int[3];

        for (int i = 0 ; i < games ; ++i) {
            Game g = new Game();
            while (true) {
                if (g.winner() is int w) {
                    won[w] += 1;
                    WriteLine($"game {i}: winner = {names[w]}");
                    break;
                }
                g.move(players[g.turn]!.move(g));
            }
        }
        WriteLine($"{names[1]} won {won[1]}, {names[2]} won {won[2]}, {won[0]} draws");
    }

    static void Main(string[] args) {
        if (args.Length == 0) {
            WriteLine("usage: tic_tac_toe <ai_player> | -g <count> <ai_1> <ai_2>");
            return;
        }
        if (args.Length > 0 && args[0] == "-g") {
            int games = int.Parse(args[1]);
            tournament(games, args[2], args[3]);
        } else {
            Player ai = player(args[0]);
            Application.Init();
            new Frame(ai).ShowAll();
            Application.Run();
        }
    }
}
